/****Classe DataBuffer, Framework Nemo, Copyright Cyril Crassin 2005****/

#ifndef DataBuffer_h
#define DataBuffer_h

#include "UtilsGL.h"

/** 
  * GPU generic data buffer.
  * Abstract VBO/PBO for fast asynchronous CPU/GPU data transfers.
  */
class DataBuffer {
protected:
	int size;
	GLuint buffId;
	GLuint lastTarget;
	GLuint lastUsage;
	GLuint lastMapAcces;
	bool mapped;
	
	void *mappedAddress;

	virtual void initialize(){
		buffId=0;
		glGenBuffersARB(1, &this->buffId);
		this->mapped=false;
		//this->lastTarget=GL_PIXEL_UNPACK_BUFFER_EXT;//GL_ARRAY_BUFFER_ARB;
		//this->lastUsage=GL_STATIC_READ_ARB;//GL_STREAM_COPY_ARB;//GL_DYNAMIC_COPY_ARB;
		this->lastMapAcces=0;
		mapped=false;
		mappedAddress=NULL;

	}

public:
	DataBuffer(int s, GLuint target=GL_ARRAY_BUFFER, GLuint usage=GL_STATIC_DRAW_ARB):lastTarget(target), lastUsage(usage){
		this->initialize();
		this->setData(NULL, s);
	}
	DataBuffer(int s, void *d, GLuint target=GL_ARRAY_BUFFER, GLuint usage=GL_STATIC_DRAW_ARB):lastTarget(target), lastUsage(usage){
		this->initialize();
		this->setData(d, s);
	}

	~DataBuffer(){
		glDeleteBuffersARB(1, &this->buffId);
	}

	/**Set GL usage of the buffer*/
	void setUsage(unsigned int usage){
		this->lastUsage=usage;
	}

	/**Set GL target of the buffer*/
	void setTarget(unsigned int target){
		this->lastTarget=target;
	}

	/**Set data into buffer*/
	void setData(void *data, int size){
		glBindBuffer(this->lastTarget, this->buffId);
		this->size=size;

		glBufferDataARB(this->lastTarget, size, data, this->lastUsage);
		glBindBuffer(this->lastTarget, 0);
	}

	/**Update a subpart of the buffer's data*/
	void setSubData(void *data, int size, int offset=0){
		if( offset<0 || offset>this->size-1 || size+offset>this->size )
			return;

		glBindBuffer(this->lastTarget, this->buffId);
		glBufferSubDataARB(this->lastTarget, offset, size, data);
	
		glBindBuffer(this->lastTarget, 0);
	}

	/**Retrieve buffer data into central memory
	 * @param	data	Central memory pointer to target space.
	 * @param	offset	Retrieved data offset into source buffer.
	 * @param	size	Retrieved data size.
	*/
	void getData(void *data, int offset=0, int size=-1){
		if(size==-1)
			size=this->size;
//		this->WaitData();
		glBindBuffer(this->lastTarget, this->buffId);
		glGetBufferSubDataARB(this->lastTarget, offset, size, data);
		glBindBuffer(this->lastTarget, 0);
	}

	/**Get data size*/
	int getSize(){
		return this->size;
	}

	/**Map data into central memory adress space
	 * @param	acces	GL access rights.
	*/
	virtual void *lock(GLuint acces=GL_READ_WRITE){
		//this->WaitData();
    
		GLuint target=this->lastTarget;

		if(this->mapped)
			if(this->lastMapAcces!=acces || target!=this->lastTarget)
				this->unlock();
			else
				return this->mappedAddress;
	    
	    
		glBindBuffer(target, this->buffId);
		void *buff=glMapBuffer(target, acces);
		glBindBuffer(target, 0);
	    
		int tmp;
		if( tmp=glGetError() ){
			std::cout<<"ERROR GPUDataArray mapping failed.\n";
			std::cout<<gluErrorString(tmp)<<"\n";
			return 0;
		}
	    
		this->lastTarget=target;
		this->lastMapAcces=acces;
		this->mapped=true;
	    
		this->mappedAddress=buff;
	    
		return buff;
	}

	/**Unmap data from central memory adress space*/
	virtual void unlock(){
		if(!this->mapped)
			return;
			
		glBindBuffer(this->lastTarget, this->buffId);
		glUnmapBuffer(this->lastTarget);
		glBindBuffer(this->lastTarget, 0);
		
		this->mapped=false;
		this->mappedAddress=NULL;
	}


	/**Use data for graphical operation*/
	virtual void *use(int offset=0){
		return ((char *)NULL + (offset));
	}

	/**Bind buffer for graphical operations
	 * @param	target	GL graphical operation type.
	*/
	virtual void bind(unsigned int target=0){
		if(target==0)
			target=this->lastTarget;
		else
			this->lastTarget=target;
	    
		glBindBuffer(target, this->buffId);
	}

	/**Unbind buffer*/
	virtual void unbind(){
		glBindBuffer(this->lastTarget, 0);
	}

};

#endif